{$MODE OBJFPC}
{$LONGSTRINGS ON}
{$APPTYPE CONSOLE}

uses SysUtils, Windows, Types;

const
  GUseEncodeFromUnicode: Boolean = True;
  GUseDecodeToUnicode: (d2uNever, d2uAlways, d2uByCodePage) = d2uByCodePage;
  GUsingLCL: Boolean = False;

function UnicodeBufToStr(const Buf: PUnicodeChar; BufLen: DWORD): string;
var
  U: UnicodeString;
begin
  SetString(U, Buf, BufLen);
  if GUseEncodeFromUnicode then
    Result := UTF8Encode(U)
  else
    Result := string(U);
end;

function StrToUnicode(const S: string): UnicodeString;
begin
  case GUseDecodeToUnicode of
    d2uAlways:
      Result := UTF8Decode(S);
    d2uNever:
      Result := UnicodeString(S);
    d2uByCodePage:
      if StringCodePage(S) = DefaultSystemCodePage then
        Result := UnicodeString(S)
      else
        Result := UTF8Decode(S);
  else
    Result := '';
  end;
end;

// Use a direct API call to exclude differences in the current implementation of TRegistry

function OpenTestKey: HKEY;
var
  LastError: LongInt;
begin
  LastError := RegOpenKeyExW(HKEY_CURRENT_USER, 'Software', 0, KEY_READ, @Result);
  if LastError <> ERROR_SUCCESS then
    RaiseLastOSError(LastError);
end;

function GetValueNames(Key: HKEY): TStringDynArray;
var
  Index: DWORD;
  Buffer: PUnicodeChar;
  BufLen, Len: DWORD;
  LastError: LongInt;
begin
  Index := 0;
  Result := nil;
  BufLen := 64;
  GetMem(Buffer, BufLen * SizeOf(UnicodeChar));
  try
    repeat
      Len := BufLen;
      LastError := RegEnumValueW(Key, Index, Buffer, Len, nil, nil, nil, nil);
      case LastError of
        ERROR_NO_MORE_ITEMS:
          Break;
        ERROR_SUCCESS:
          begin
            SetLength(Result, Length(Result) + 1);
            Result[High(Result)] := UnicodeBufToStr(Buffer, Len);
          end;
        ERROR_MORE_DATA:
          begin
            Inc(BufLen, BufLen);
            ReAllocMem(Buffer, BufLen * SizeOf(UnicodeChar));
            Continue;
          end;
      end;
      Inc(Index);
    until False;
  finally
    FreeMem(Buffer);
  end;
end;

function ReadValue(Key: HKEY; const Name: string): string;
var
  LastError: LongInt;
  U: UnicodeString;
  Buffer: PUnicodeChar;
  BufSize: DWORD;
  DataType: DWORD;
begin
  U := StrToUnicode(Name);
  BufSize := 0;
  LastError := RegQueryValueExW(Key, Pointer(U), nil, @DataType, nil, @BufSize);
  if LastError <> ERROR_SUCCESS then
    if LastError = ERROR_FILE_NOT_FOUND then
      Exit('Error: name ' + Name + ' not found')
    else
      RaiseLastOSError(LastError);
  if not (DataType in [REG_SZ, REG_EXPAND_SZ, REG_MULTI_SZ]) or (BufSize = 0) then
    Exit('');
  GetMem(Buffer, BufSize);
  try
    LastError := RegQueryValueExW(Key, Pointer(U), nil, nil, PByte(Buffer), @BufSize);
    if LastError <> ERROR_SUCCESS then
      RaiseLastOSError(LastError);
    if Buffer[BufSize div 2 - 1] = #0 then
      Dec(BufSize, SizeOf(UnicodeChar));
    Result := UnicodeBufToStr(Buffer, BufSize div SizeOf(UnicodeChar));
    if DataType = REG_MULTI_SZ then
      Result := StringReplace(Result, #0, LineEnding, [rfReplaceAll]);
  finally
    FreeMem(Buffer);
  end;
end;

procedure PrintString(const S: string);
begin
  // To exclude console font or converting ot OEM limitation duplicate to Debug
  OutputDebugStringA(PChar(S));
  Writeln(S);
end;

var
  L: TStringDynArray;
  S: string;
  Key: HKEY;
begin
  //GUsingLCL := True;
  //GUseEncodeFromUnicode := False;
  //GUseDecodeToUnicode := d2uAlways;
  if GUsingLCL then
    SetMultiByteConversionCodePage(CP_UTF8);
  try
    Key := OpenTestKey;
    try
      PrintString('[----Get value Names----]');
      L := GetValueNames(Key);
      for S in L do
        PrintString(S);
      PrintString('');
      PrintString('[----Read value by name from CONST----]');
      S := 'très tôt'; // Use string in user ACP!
      PrintString(S + '=' + ReadValue(Key, S));
      PrintString('');
      PrintString('[----Read values by name from RESULT----]');
      for S in L do
        PrintString(S + '=' + ReadValue(Key, S));
      PrintString('[----Read value by name from VAR----]');
      Write('Input value name:');
      Readln(S);
      PrintString(S + '=' + ReadValue(Key, S));
    finally
      RegCloseKey(Key);
    end;
  except
    on E: Exception do
      PrintString('Error: ' + E.Message);
  end;
  Readln;
end.

